/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= */
/*
 *  It will run on any POSIX-compliant machine.
 *
 *  Created 18-DEC-1998       Richard B. Johnson   Analogic Corporation
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <malloc.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <wait.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <netdb.h>
#include <mailer.h>

static void quit(int sig)
{
    (void)fprintf(stdout, "Aborted via ^%c\n", ('A' + sig));
    (void)fflush(stdout);
    exit(EXIT_SUCCESS);
}

static void reaper(int unused)
{
    show("Catch child status");
    while(wait3(&unused, WNOHANG, NULL) > 0)
        ;
}

static const char esmtp[]=" ESMTP Sendmail 8.11.0Beta3 ";
static const char fu[]="Kindly remove this mail port from your list "\
                       "of addresses to bombard with SPAM......";

extern void execute_command(const char *cmd);

int main(int unused, char *argv[])
{
    char *buffer;
    FILE *file;
    FILE *ipchains;
    char *prompt, *cp;
    struct hostent *host;                    /* Server address pointer  */
    struct hostent *client;
    struct sockaddr_in cli_addr;             /* Client address struct   */
    struct sockaddr_in ip_sock;              /* IP socket address       */
    struct sigaction sa;
    struct timeval tv;
    struct timezone tz;
    unsigned int addr;
    int s, t, i, fd, len;

    tzset();
    show("Standby...");
    if((file = fopen(PATH_SPAMLOG, "w")) == NULL)
       ERRORS(errno); 
    if((buffer = CP malloc(BUF_LEN * sizeof(char))) == NULL)
        ERRORS(errno);
    prompt = buffer + (2 * MAXHOSTNAMELEN);
/*
 *  Set the wait-for-status trap for all child processes.
 */
    if(sigaction(SIGCHLD, NULL, &sa) < 0)
        ERRORS(errno);
    sa.sa_flags = SA_INTERRUPT|SA_RESTART;
    sa.sa_handler = reaper;
    if(sigaction(SIGCHLD, &sa, NULL) < 0)
        ERRORS(errno);
/*
 *  Set up for orderly shutdown from ^C.
 */
    if(sigaction(SIGINT, NULL, &sa) < 0)
        ERRORS(errno);
    sa.sa_flags = SA_INTERRUPT;
    sa.sa_handler = quit;
    if(sigaction(SIGINT, &sa, NULL) < 0)
        ERRORS(errno);
/*
 *  Set up a socket to listen for connections from the client.
 */
    if(gethostname(buffer, BUF_LEN) < 0)
        ERRORS(errno);
    if((host = gethostbyname(buffer)) == NULL)
        ERRORS(errno);
    if((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
        ERRORS(errno);
    i = 1; 
    if((setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i))) < 0)
        ERRORS(errno);
    memset(&ip_sock, 0x00, sizeof(ip_sock));
    ip_sock.sin_addr.s_addr = *((long *)host->h_addr);
    ip_sock.sin_family = AF_INET;
    ip_sock.sin_port = htons(PORT);
    if((bind(s, SSAP &ip_sock, sizeof(ip_sock))) < 0)
        ERRORS(errno);
    if((listen(s, BACKLOG)) < 0)
        ERRORS(errno);
/*
 *  Okay. Socket is set up, now accept connections <forever>.
 */
    for(;;)
    {
        fd = -1;
        show("Waiting for a connection");
        i = sizeof(ip_sock.sin_addr);
        if((t = accept(s, SSAP &ip_sock, &i)) < 0)
            ERRORS(errno);
        i = sizeof(cli_addr);
        show("Get address of client");
        if(getpeername(t, SSAP &cli_addr, &i) < 0)
            ERRORS(errno);
        show("Get name of client");
        addr = (unsigned long) cli_addr.sin_addr.s_addr;
/*
 *  This needs to be opened, written, then closed or else an attempt
 *  to execute it returns EBUSY.
 */
        if(access(PATH_DENY, F_OK))
        {
            if((ipchains = fopen(PATH_DENY, "a+")) == NULL)
                ERRORS(errno);
            (void)fprintf(ipchains, "#!/bin/sh\n");
        }
        else
        {
            if((ipchains = fopen(PATH_DENY, "a+")) == NULL)
                ERRORS(errno);
        }
        (void)fprintf(ipchains,
            "ipchains -A input -i eth0 -s %u.%u.%u.%u -p tcp -y -j DENY\n",
                           addr & 0xff,
                           (addr >> 0x08) & 0xff,
                           (addr >> 0x10) & 0xff,
                           (addr >> 0x18) & 0xff);
        (void)fflush(ipchains);
        (void)fclose(ipchains);
        (void) chmod("deny.sh", 0744); 
        (void)system("./deny.sh");
        if((client = gethostbyaddr(CP &cli_addr.sin_addr, 
                          sizeof(cli_addr.sin_addr), AF_INET)) == NULL)
        {
            (void)fprintf(file, "Unresolvable address %u.%u.%u.%u\n",
                           addr & 0xff,
                           (addr >> 0x08) & 0xff,
                           (addr >> 0x10) & 0xff,
                           (addr >> 0x18) & 0xff);
            (void)fflush(file);
            (void)close(t);
            continue;
        }
        show("Client connected");
        (void)sprintf(buffer, "./resp %s", client->h_name);
        execute_command(buffer);
        switch(fork())
        {
        case 0:
            show("Server task child");
            if(close(s) < 0)
                ERRORS(errno);
            (void)sprintf(argv[0], "Serving %s", client->h_name);
            (void)gettimeofday(&tv, &tz);
            (void)fprintf(file, "Spammer %s (%u.%u.%u.%u) connected at %s",
                           client->h_name,
                           addr & 0xff,
                           (addr >> 0x08) & 0xff,
                           (addr >> 0x10) & 0xff,
                           (addr >> 0x18) & 0xff,
                           asctime(localtime(&tv.tv_sec)));
            fflush(file);
            if(sigaction(SIGINT, NULL, &sa) < 0)
                ERRORS(errno);
            sa.sa_flags = SA_INTERRUPT;
            sa.sa_handler = SIG_DFL;
            if(sigaction(SIGINT, &sa, NULL) < 0)
                ERRORS(errno);
            show("Main server loop");
            show("Sending prompt");
            strcpy(prompt, "220 ");
            gethostname(buffer, MAXHOSTNAMELEN);
            if((cp = strchr(buffer, '\n')) != NULL)
                *cp = (char) 0x00;
            strcat(prompt, buffer);
            strcat(prompt, ".");
            getdomainname(buffer, MAXHOSTNAMELEN);
            if((cp = strchr(buffer, '\n')) != NULL)
                *cp = (char) 0x00;
            strcat(prompt, buffer);
            strcat(prompt, esmtp);
            (void)gettimeofday(&tv, &tz);
            strcat(prompt, asctime(localtime(&tv.tv_sec)));
            if((cp = strchr(prompt, '\n')) != NULL)
                *cp = (char) 0x00;
            sprintf(buffer, " -%d\n", tz.tz_minuteswest);
            strcat(prompt, buffer); 
            (void)write(t, prompt, strlen(prompt)); 
            show("Waiting to read");
            if((i = read(t, buffer, BUF_LEN)) <= 0)
            {
                (void)gettimeofday(&tv, &tz);
                (void)fprintf(file, "Spammer %s disconnected at %s",
                                 client->h_name,
                                 asctime(localtime(&tv.tv_sec)));
                (void)fflush(file);
                (void)close(t);
                exit(EXIT_FAILURE);
            }
            cp = buffer;
            for(i=0; i< BUF_LEN/(sizeof(fu)-1); i++)
            {
                memcpy(cp, fu, sizeof(fu)-1);
                cp += sizeof(fu);
            } 
            for(;;)
            {
                len = BUF_LEN;
                cp = buffer;
                while(len)
                {
                    if((i = write(t, cp, len)) <= 0)
                    {
                        (void)gettimeofday(&tv, &tz);
                        (void)fprintf(file, "Spammer %s disconnected at %s",
                                         client->h_name,
                                         asctime(localtime(&tv.tv_sec)));
                        fflush(file);
                        (void)close(t);
                        exit(EXIT_FAILURE);
                    }
                    len -= i;
                    cp  += i;
                    usleep(20000);             /* Be kind to LAN        */
                }
                sleep(1);
            }
            break;
        case -1:                               /*  Means fork failed    */ 
            ERRORS(errno);
            break;
        default:                               /* Parent continues here */
            (void)close(t);                    /* Only child has open   */
            (void)usleep(1000);                /* Throttle connections  */
            break;
        }
    } 
    fclose(file);
    return 0;
}
/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= */
/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= */
/* -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= */
